/********************************************************************************/
/*    H   H Y   Y TTTTT EEEEE  CCC       HYTEC ELECTRONICS LTD                  */
/*    H   H  Y Y    T   E     C          5 Cradock Road,                        */
/*    HHHHH   Y     T   EEE   C          Reading, Berks.    Tel: 0118 9757770   */
/*    H   H   Y     T   E     C          RG2 0JT            Fax: 0118 9757566   */
/*    H   H   Y     T   EEEEE  CCC       Web: www.hytec-electronics.co.uk       */
/********************************************************************************/

/********************************************************************************/
/*                                                                              */
/*  devHy8506.c - Device Support for the Hytec 8506IP cards                     */
/*                                                                              */
/*                               .                                              */
/*                                                                              */
/********************************************************************************/
/*                                                                              */
/*  Revision history: (comment and initial revisions)                           */
/*                                                                              */
/*  vers.       revised         modified by                                     */
/*  -----       -----------     -------------------------                       */
/*  1.1         15-DEC-2010     Jim Chen Hytec Electronics Ltd                  */
/*	                        Driver support for a Hytec Hy8506 IP            	*/
/*			        		module. Initial Version.                        	*/
/*                                                                              */
/********************************************************************************/
/*#include <vxWorks.h>*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <dbCommon.h>
#include <biRecord.h>
#include <boRecord.h>
#include <mbbiRecord.h>
#include <mbboRecord.h>
#include <mbbiDirectRecord.h>
#include <mbboDirectRecord.h>

#include <dbDefs.h>
#include <dbAccess.h>
#include <recSup.h>
#include <devSup.h>
#include <dbScan.h>
#include <alarm.h>
#include <recGbl.h>

#include <registryFunction.h>
#include <iocsh.h>
#include <epicsExport.h>

#define CONVERT		0
#define NO_CONVERT	2

int Hy8506getioscanpvt(const int cardnum, IOSCANPVT *scanpvt);

static long init(void);
static long init_bi(struct biRecord *pbi);
static long init_bo(struct boRecord *pbo);
static long init_mbbi(struct mbbiRecord *pmbbi);
static long init_mbbo(struct mbboRecord *pmbbo);
static long init_mbbiDirect(struct mbbiDirectRecord *pmbbiDirect);
static long init_mbboDirect(struct mbboDirectRecord *pmbboDirect);

static long  read_bi(struct biRecord *pbi);
static long write_bo(struct boRecord *pbo);
static long  read_mbbi(struct mbbiRecord *pmbbi);
static long write_mbbo(struct mbboRecord *pmbbo);
static long  read_mbbiDirect(struct mbbiDirectRecord *pmbbiDirect);
static long write_mbboDirect(struct mbboDirectRecord *pmbboDirect);

static long bi_get_ioint_info(int cmd, struct biRecord *p, IOSCANPVT *pvt);
static long mbbi_get_ioint_info(int cmd, struct mbbiRecord *p, IOSCANPVT *pvt);
static long mbbiDirect_get_ioint_info(int cmd, struct mbbiDirectRecord *p, IOSCANPVT *pvt);

long drvHy8506_read (const int cardnum, int port, unsigned short *value);
long drvHy8506_write(const int cardnum, int port, const unsigned short mask, unsigned short value);

struct dev_sup
{
	long		number;
	DEVSUPFUN	report;
	DEVSUPFUN	init;
	DEVSUPFUN	init_record;
	DEVSUPFUN	get_ioint_info;
	DEVSUPFUN	read_write;
};

struct dev_sup devBoHy8506 =
{
	5,
	NULL,
	init,
	init_bo,
	NULL,
	write_bo
};

struct dev_sup devBiHy8506 =
{
	5,
	NULL,
	NULL,
	init_bi,
	bi_get_ioint_info,
	read_bi
};

struct dev_sup devMbboHy8506 =
{
	5,
	NULL,
	init,
	init_mbbo,
	NULL,
	write_mbbo
};

struct dev_sup devMbbiHy8506 =
{
	5,
	NULL,
	NULL,
	init_mbbi,
	mbbi_get_ioint_info,
	read_mbbi
};

struct dev_sup devMbboDirectHy8506 =
{
	5,
	NULL,
	init,
	init_mbboDirect,
	NULL,
	write_mbboDirect
};

struct dev_sup devMbbiDirectHy8506 =
{
	5,
	NULL,
	NULL,
	init_mbbiDirect,
	mbbiDirect_get_ioint_info,
	read_mbbiDirect
};

/*
	No device specific initialisation required
*/

static long init(void)
{
	return (0);
}

/* bi ****************************************************************/

static long init_bi(struct biRecord *pbi)
{
	struct vmeio *pvmeio;
	
/* bi.inp must be an VME_IO */

	switch (pbi->inp.type)
	{
		case (VME_IO):
		{
			pvmeio = (struct vmeio *)&(pbi->inp.value);
			
			pbi->mask = 1;
			pbi->mask <<= pvmeio->signal;
			break;
		}
		
		default:
		{
			recGblRecordError(S_db_badField, (void *) pbi, "devHy8506 (init_record) Illegal INP field");
			return(S_db_badField);
		}
	}

	return (0);
}

static long read_bi(struct biRecord *pbi)
{
	struct vmeio *pvmeio;
	long status;
	unsigned short value;
	int port;
	
	pvmeio = (struct vmeio *) &(pbi->inp.value);
	
	port = (int)((*(pvmeio->parm) & 0xFF) - 0x30);	
	status = drvHy8506_read(pvmeio->card, port, &value);
	
	if (status == 0)
	{
		pbi->rval = value & pbi->mask;
		return (CONVERT);
	}

	recGblSetSevr(pbi, READ_ALARM, INVALID_ALARM);	
	return (NO_CONVERT);
}

static long bi_get_ioint_info(int cmd, struct biRecord *p, IOSCANPVT *pvt)
{
	Hy8506getioscanpvt(p->inp.value.vmeio.card, pvt);

	return (0);
}

/* bo ****************************************************************/

static long init_bo(struct boRecord *pbo)
{
	struct vmeio *pvmeio;

	pvmeio = (struct vmeio *)&(pbo->out.value);

/* bo.out must be an VME_IO */

	switch (pbo->out.type)
	{
		case (VME_IO):
		{
			pbo->mask = 1;
			pbo->mask <<= pvmeio->signal;

			break;
		}
		
		default:
		{
			recGblRecordError(S_db_badField, (void *) pbo, "devHy8506 (init_record) Illegal OUT field");
			return(S_db_badField);
		}
	}

	return (0);
}

static long write_bo(struct boRecord *pbo)
{
	struct vmeio *pvmeio;
	long status;
	int port;

	pvmeio = (struct vmeio *) &(pbo->out.value);

/*	printf("devHy8506_write_bo 0x%04x\n", (int) pbo->rval); */
	
	port = (int)((*(pvmeio->parm) & 0xFF) - 0x30);	
	status = drvHy8506_write(pvmeio->card, port, pbo->mask, (unsigned short) pbo->rval);
	
	if (status != 0)
	{
		recGblSetSevr(pbo, WRITE_ALARM, INVALID_ALARM);
	}
	
	return (0);
}

/* mbbi **************************************************************/

/*
	for mbbi mask is initialised with 2^(no bits) - 1
*/

static long init_mbbi(struct mbbiRecord *pmbbi)
{
	
/* mbbi.inp must be an VME_IO */

	switch (pmbbi->inp.type)
	{
		case (VME_IO):
		{
			pmbbi->mask <<= pmbbi->shft;
			break;
		}
		
		default:
		{
			recGblRecordError(S_db_badField, (void *) pmbbi, "devHy8506 (init_record) Illegal INP field");
			return(S_db_badField);
		}
	}

	return (0);
}

static long read_mbbi(struct mbbiRecord *pmbbi)
{
	struct vmeio *pvmeio;
	long status;
	unsigned short value;
	int port;

	pvmeio = (struct vmeio *) &(pmbbi->inp.value);
	
	port = (int)((*(pvmeio->parm) & 0xFF) - 0x30);	
	status = drvHy8506_read(pvmeio->card, port, &value);
	
	if (status == 0)
	{
		pmbbi->rval = value & pmbbi->mask;
		return (CONVERT);
	}

	recGblSetSevr(pmbbi, READ_ALARM, INVALID_ALARM);
	return (NO_CONVERT);
}

static long mbbi_get_ioint_info(int cmd, struct mbbiRecord *p, IOSCANPVT *pvt)
{
	Hy8506getioscanpvt(p->inp.value.vmeio.card, pvt);

	return (0);
}

/* mbbo **************************************************************/

static long init_mbbo(struct mbboRecord *pmbbo)
{
	
/* mbbo.out must be an VME_IO */

	switch (pmbbo->out.type)
	{
		case (VME_IO):
		{
			pmbbo->mask <<= pmbbo->shft;
			break;
		}
		
		default:
		{
			recGblRecordError(S_db_badField, (void *) pmbbo, "devHy8506 (init_record) Illegal OUT field");
			return(S_db_badField);
		}
	}

	return (0);
}

static long write_mbbo(struct mbboRecord *pmbbo)
{
	struct vmeio *pvmeio;
	long status;
	int port;

	pvmeio = (struct vmeio *) &(pmbbo->out.value);

/*	printf("devHy8506_write_mbbo   rval 0x%04x   val 0x%04x\n", (int) pmbbo->rval, pmbbo->val); */
	
	port = (int)((*(pvmeio->parm) & 0xFF) - 0x30);	
	status = drvHy8506_write(pvmeio->card, port, pmbbo->mask, (unsigned short) pmbbo->rval);
	
	if (status != 0)
	{
		recGblSetSevr(pmbbo, WRITE_ALARM, INVALID_ALARM);
	}
	
	return (0);
}

/* mbbiDirect ********************************************************/

/*
	for mbbiDirect mask is initialised with 2^(no bits) - 1
*/

static long init_mbbiDirect(struct mbbiDirectRecord *pmbbiDirect)
{
	
/* mbbiDirect.inp must be an VME_IO */

	switch (pmbbiDirect->inp.type)
	{
		case (VME_IO):
		{
			pmbbiDirect->mask <<= pmbbiDirect->shft;
			break;
		}
		
		default:
		{
			recGblRecordError(S_db_badField, (void *) pmbbiDirect, "devHy8506 (init_record) Illegal INP field");
			return(S_db_badField);
		}
	}

	return (0);
}

static long read_mbbiDirect(struct mbbiDirectRecord *pmbbiDirect)
{
	struct vmeio *pvmeio;
	long status;
	unsigned short value;
	int port;

	pvmeio = (struct vmeio *) &(pmbbiDirect->inp.value);
	
	port = (int)((*(pvmeio->parm) & 0xFF) - 0x30);	
	status = drvHy8506_read(pvmeio->card, port, &value);
	
	if (status == 0)
	{
		pmbbiDirect->rval = value & pmbbiDirect->mask;
		return (CONVERT);
	}

	recGblSetSevr(pmbbiDirect, READ_ALARM, INVALID_ALARM);
	return (NO_CONVERT);
}

static long mbbiDirect_get_ioint_info(int cmd, struct mbbiDirectRecord *p, IOSCANPVT *pvt)
{
	Hy8506getioscanpvt(p->inp.value.vmeio.card, pvt);

	return (0);
}

/* mbboDirect ********************************************************/

static long init_mbboDirect(struct mbboDirectRecord *pmbboDirect)
{
	
/* mbboDirect.out must be an VME_IO */

	switch (pmbboDirect->out.type)
	{
		case (VME_IO):
		{
			pmbboDirect->mask <<= pmbboDirect->shft;
			break;
		}
		
		default:
		{
			recGblRecordError(S_db_badField, (void *) pmbboDirect, "devHy8506 (init_record) Illegal OUT field");
			return(S_db_badField);
		}
	}

	return (0);
}

static long write_mbboDirect(struct mbboDirectRecord *pmbboDirect)
{
	struct vmeio *pvmeio;
	long status;
	int port;

	pvmeio = (struct vmeio *) &(pmbboDirect->out.value);

	port = (int)((*(pvmeio->parm) & 0xFF) - 0x30);	
	status = drvHy8506_write(pvmeio->card, port, pmbboDirect->mask, (unsigned short) pmbboDirect->rval);
	
	if (status != 0)
	{
		recGblSetSevr(pmbboDirect, WRITE_ALARM, INVALID_ALARM);
	}
	
	return (0);
}


epicsExportAddress(dset, devBoHy8506);
epicsExportAddress(dset, devBiHy8506);
epicsExportAddress(dset, devMbboHy8506);
epicsExportAddress(dset, devMbbiHy8506);
epicsExportAddress(dset, devMbboDirectHy8506);
epicsExportAddress(dset, devMbbiDirectHy8506);
