/********************************************************************************/
/*    H   H Y   Y TTTTT EEEEE  CCC       HYTEC ELECTRONICS LTD                  */
/*    H   H  Y Y    T   E     C          5 Cradock Road,                        */
/*    HHHHH   Y     T   EEE   C          Reading, Berks.    Tel: 0118 9757770   */
/*    H   H   Y     T   E     C          RG2 0JT            Fax: 0118 9757566   */
/*    H   H   Y     T   EEEEE  CCC       Web: www.hytec-electronics.co.uk       */
/********************************************************************************/

/********************************************************************************/
/*                                                                              */
/*  drvHy8506.c - Driver Support for the Hytec 8506IP cards                     */
/*                                                                              */
/*                               .                                              */
/*                                                                              */
/********************************************************************************/
/*                                                                              */
/*  Revision history: (comment and initial revisions)                           */
/*                                                                              */
/*  vers.       revised         modified by                                     */
/*  -----       -----------     -------------------------                       */
/*  1.1         15-DEC-2010     Jim Chen Hytec Electronics Ltd                  */
/*	                        Driver support for a Hytec Hy8506 IP            	*/
/*			        		module. Initial Version.                        	*/
/*                                                                              */
/********************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <devLib.h>
#include <dbScan.h>
#include <drvSup.h>

#include "Hy8506.h"
#include <drvIpac.h>

#include <epicsExit.h>
#include <registryFunction.h>
#include <iocsh.h>
#include <epicsExport.h>
#include <dbDefs.h>

#undef DEBUG

/* io structure */

struct ioregs
{

/* writable I/O registers */

	unsigned short LKC;	/* last known change */
	unsigned short CSR;	/* control / status register */
	unsigned short IMR;	/* interrupt mask register */
	unsigned short DBR;	/* debounce register */
	unsigned short PSR;	/* pulse select register */
	unsigned short PPR;	/* pulse parameter register */
	unsigned short VEC;	/* interrupt vector */
	unsigned short NOTUSED1;	/* not used */
};

/* structure to support multiple IP modules */

struct cardinfo
{
/* next structure */
	struct cardinfo *next;

/* for EPICS I/O interrupt */
	IOSCANPVT ioscanpvt;

/* config details */
	
	unsigned short cardnum;		/* EPICS card number */
	unsigned short carrier;		/* carrier card number */
	unsigned short ipslot;		/* IP slot number */
	unsigned short init;		/* module initialised */

	int invert[3];				/* If set all input/output bits are inverted. */
	unsigned short vector[3];	/* IP interrupt vector */
	unsigned short intmask[3];	/* interrupt mask */
	unsigned short debrate[3];	/* debounce rate */	
	volatile struct ioregs *regs[3];		/* registers in I/O space on card */
	struct ioregs restore[3];	/* registers to restore */
};

static long REPORT(int level);
static long INIT(void);
static long LIST(void);

static int dumpcard(struct cardinfo *card);

long drvHy8506_read (const int cardnum, const int port, unsigned short *value);
long drvHy8506_write(const int cardnum, const int port, const unsigned short mask, unsigned short value);

/* our single linked list of cardinfo structures */

static struct cardinfo *cardlist = NULL;

struct drv_sup
{
	long		number;
	DRVSUPFUN	report;
	DRVSUPFUN	init;
};

struct drv_sup drvHy8506 =
{
	2,
	REPORT,
	INIT
};

static long drvHy8506Report(int level)
{
	int i;

	switch (level)
	{
		case 0:		;
		case 1:		LIST();
				break;
		case 2:
		{
			struct cardinfo *card = cardlist;
			
		/* display the registers */
		
			while (card != NULL)
			{		
				for(i=0; i<3; i++)
				{
					printf("\tLKC: %04x\tCSR: %04x\n", card->regs[i]->LKC, card->regs[i]->CSR);
					printf("\tIMR: %04x\tDBR: %04x\n", card->regs[i]->IMR, card->regs[i]->DBR);
					printf("\tPSR: %04x\tPPR: %04x\n", card->regs[i]->PSR, card->regs[i]->PPR);
				}
				printf("\tVEC: %04x CardNo: %d\n\n", card->regs[0]->VEC, card->cardnum);
				
				card = card->next;
			}
			
			break;
		}
	
		default:
		{
			break;
		}
	}
	
	return (0);
}


/*
	device layer masks out all bits defined by the signal number and the shift
*/

long drvHy8506_read(const int cardnum, const int port, unsigned short *value)
{
	struct cardinfo *card = cardlist;
	
	while ((card != NULL) && (card->cardnum != cardnum))
	{
		card = card->next;
	}
	
/* card wasn't found */

	if (card == NULL)
	{
		return (S_drv_badParam);
	}

/* read inputs, invert if required */
	
	*value = card->regs[port]->LKC;
	if (card->invert[port])
		*value = ~ *value;

#ifdef DEBUG
	printf("drvHy8506_read - card %2d  port %d val 0x%04x\n", cardnum, port, *value);
#endif	

	return (S_drv_OK);
}

/*
	Set individual bits

*/

long drvHy8506_write(const int cardnum, const int port, const unsigned short mask, unsigned short value)
{
	struct cardinfo *card = cardlist;

	while ((card != NULL) && (card->cardnum != cardnum))
	{
		card = card->next;
	}
	
/* card wasn't found */

	if (card == NULL)
	{
		return (S_drv_badParam);
	}

/* write outputs, inverting if requested */

	if (card->invert[port])
		value = ~value;
	if (mask == 0xffff)
	{
		card->regs[port]->LKC = card->restore[port].LKC = value;
	}
	else
	{
		card->restore[port].LKC = card->regs[port]->LKC = (card->regs[port]->LKC & ~mask) | (value & mask);
	}
	
#ifdef DEBUG
	printf("drvHy8506_write - card %2d  port %d mask 0x%04x  val 0x%04x LKC %x restLKC %x\n", cardnum, port, mask, value, card->regs[port]->LKC, card->restore[port].LKC);
#endif
	
	return (S_drv_OK);
}

/*********************************************************************/

#define PR_ID    0x06
#define PR_MODN  0x0a
#define PR_REVN  0x0c
#define PR_DRID1 0x10
#define PR_DRID2 0x12
#define PR_FLAGS 0x14
#define PR_BYTES 0x16
#define PR_SERN  0x1a
#define HYTECID  0x00800300

/*
	check that the correct IP module is installed
*/

static int checkprom(volatile struct ioregs *ptrb)
{
	int ret = 0;
	unsigned int manid;
	unsigned short model, rev, serial;
	char header[16], tmp;
	char * ptr = (char*)ptrb + PROM_OFFS;
	
	if(strcmp("VITA4 ", (char*)ptr) != 0) 
		ret += 1;
	memcpy(header, ptr, 2);
	tmp = header[1];
	header[1] = header[0];
	header[0] = tmp;
	memcpy(&header[2], ptr+2, 2);
	tmp = header[3];
	header[3] = header[2];
	header[2] = tmp;	
	memcpy(&header[4], ptr+4, 2);
	header[6]=0;

	manid  = *((short *)   (ptr +  8))  + ((*((short *)(ptr +  6))) << 16);
	model  = *((short *) (ptr + 10));
	rev    = *((short *) (ptr + 12));
	serial = *((short *) (ptr + 26));

	printf("%s: %s  model: %04X  manid: %08X  rev: %04X  serial: %d\n", DRIVER, header, model, manid, rev, serial);

	if (manid != MANID) ret += 2;
	if (model != MODEL) ret += 4;
	if (rev    < REV)   ret += 8;

	return (ret);
}

/*
	write data back to the IP module
	
	The order is important: CSR should be written before LKC so that the card is initialised for
	input or output before the output is set.
*/

static int dumpcard(struct cardinfo *card)
{
	int i;

	for(i=0; i<3; i++)
	{
	/*	card->regs->LKC = card->restore.LKC;	*/		/* we don't need to do this. imagine it is input mode */
		card->regs[i]->IMR = card->restore[i].IMR;
		card->regs[i]->DBR = card->restore[i].DBR;
		card->regs[i]->PSR = card->restore[i].PSR;
		card->regs[i]->PPR = card->restore[i].PPR;
		card->regs[i]->VEC = card->restore[i].VEC;

		card->regs[i]->CSR = card->restore[i].CSR & 0xBFFF;	/* disable interrupt at start */
		card->regs[i]->CSR = card->restore[i].CSR & 0x3FFF;	/* clear interrupt as well */
		card->regs[i]->CSR = card->restore[i].CSR;			/* enable interrupt */
	}

	return OK;
}

/*
	interrupt service routine
*/

static void drvISR(const int cardnum)
{
	struct cardinfo *card = cardlist;

	while ((card != NULL) && (card->cardnum != cardnum))
	{
		card = card->next;
	}
	
/* card wasn't found */

	if (card == NULL)
	{
		return;
	}
	
/* clear the interrupt */
	card->regs[0]->CSR &= 0x7fff;
	card->regs[1]->CSR &= 0x7fff;
	card->regs[2]->CSR &= 0x7fff;

/* inform EPICS */
	scanIoRequest(card->ioscanpvt);

/* Note, the interrupt enable might need to be done in ioscanpvt */
	ipmIrqCmd(card->carrier, card->ipslot, 0, ipac_irqEnable);	
}

/*
	list the cards and ipslots
*/

static long LIST(void)
{
	struct cardinfo *card = cardlist;

	while (card != NULL)
	{
		printf("%s:  card: %d  carrier: %2d  ip: %d  %s\n", DRIVER, card->cardnum, card->carrier, card->ipslot, (card->init) ? "configured" : "");
		card = card->next;
	}
	
	return (0);
}

/*
	This CONFIGURE initialises the general settings for all three ports. The user calls the 
	Hy8506PortConfig function to configure the nitty-gritty things.
	
	cardnum		EPICS card number (User defines)
	carrier		carrier number returned by ipacAddxxxx
	ipslot		ipslot 0-5
	vector		interrupt vector

*/

int CONFIGURE(	const int cardnum, const int carrier, const int ipslot, const int vector)
{
	struct cardinfo *card = cardlist;

/* check parameters */

	while (card != NULL)
	{
		if (card->cardnum == cardnum) 
		{
			printf("%s: card %d already defined.\n", DRIVER, cardnum);
			card->init = 0xffff;
			return (-1);
		}
		else
		{
			if ((card->carrier == carrier) && (card->ipslot == ipslot))
			{
				printf("%s: Carrier and IP slot already defined.\n", DRIVER);
				card->init = 0xffff;
				return (-1);
			}
		}
			
		card = card->next;
	}

	/* allocate structure and initialise */

	card = calloc(1, sizeof(struct cardinfo));
	
	if (card == NULL)
	{
		printf("%s: Memory allocation error.\n", DRIVER);
		card->init = 0xffff;
		return (-1);
	}

	if ((ipslot < 0) || (ipslot > 6))
	{
		printf("%s: IP slot number %d invalid\n", DRIVER, ipslot);
		card->init = 0xffff;
		return (-1);
	}

	card->next = cardlist;
	cardlist   = card;

	card->cardnum = cardnum;
	card->carrier = carrier;
	card->ipslot  = ipslot;
	card->vector[0]  = vector & 0xff;		/* interrupt vector for all three ports*/
	card->vector[1]  = vector & 0xff;		
	card->vector[2]  = vector & 0xff;		
				
	printf("%s:  card: %d  carrier: %2d  ip: %d  int: %d\n", DRIVER, cardnum, carrier, ipslot, vector);
		
	return (0);
}


/*
	Configure the port. Checks that the parameters are correct and that the card hasn't already 
	been configured. EPICS will call the driver initialisation function later which will do the 
	real configuration.

	cardnum		EPICS card number (User defines)

	port		port number. must be 0, 1 or 2

	debrate		debounce rate	0 = none
					1 = 100Hz
					2 = 200Hz
					3 = 500Hz
					4 = 1kHz

	debmask		select input bits to debounce 0x0000 - 0xFFFF
	
	scanrate	input scan rate	0 = 1kHz
					1 = 10kHz
					2 = 100kHz
					3 = 1MHz
	dir		bit0 = 0: all 16 channels as inputs
			bit0 = 1: all 16 channels as outputs

			Bit1 encode whether the input or output
			values should be inverted.
			bit1 = 0: No inversion
			bit1 = 1: Invert all input/output bits
				
	intmask		select bits to generate interrupts	0x0000 - 0xFFFF

	pwidth		pulse width	0 = 1 msec
					1 = 10 msec
					2 = 100 msec
					3 = 1 sec
					4 = 2 sec
					5 = 5 sec
					6 = 10 sec
					7 = 20 sec
					8 = 50 sec
					9 = 100 sec

	pmask		select bits to pulse on output		0x0000 - 0xFFFF
	
	clock		0 = internal, 1 = external

*/
#define INVERT_BIT	2

int Hy8506PortConfig(int cardnum, int port, int debrate, 
		int pwidth, int scanrate, int dir, int clock, 
		int debmask, int pmask, int intmask)
{
	struct cardinfo *card = cardlist;
	const char *dirs[] = { "in", "out"};
	const char *scans[] = { "1k", "10k", "100k", "1M"};
	const char *bounce[] = { "100", "200", "500", "1k"};
	const char *widths[] = { "1 msec", "10 msec", "100 msec", "1 sec", "2 sec", "5 sec", "10 sec", "20 sec", "50 sec", "100 sec"};

	while ((card != NULL) && (card->cardnum != cardnum))
	{
		card = card->next;
	}
	
/* card wasn't found */

	if (card == NULL)
	{
		printf("%s:   card number %d not found\n", DRIVER, cardnum);
		card->init = 0xffff;
		return (-1);
	}
	
/* check parameters */

	if ((port < 0) || (port > 2))
	{
		printf("%s: port number %d invalid\n", DRIVER, port);
		card->init = 0xffff;
		return (-1);
	}

	if ((debrate < 0) || (debrate > 4))
	{
		printf("%s: debounce rate parameter %d invalid\n", DRIVER, debrate);
		card->init = 0xffff;
		return (-1);
	}

	if ((scanrate < 0) || (scanrate > 3))
	{
		printf("%s: scan rate parameter %d invalid\n", DRIVER, scanrate);
		card->init = 0xffff;
		return (-1);
	}

	if ((dir < 0) || (dir > 3))
	{
		printf("%s: direction parameter %d invalid \n", DRIVER, dir);
		card->init = 0xffff;
		return (-1);
	}

	if ((pwidth < 0) || (pwidth > 9))
	{
		printf("%s: pulse width parameter %d invalid\n", DRIVER, pwidth);
		card->init = 0xffff;
		return (-1);
	}

	if ((clock != 0) && (clock != 1))
	{
		printf("%s: clock parameter %d invalid\n", DRIVER, clock);
		card->init = 0xffff;
		return (-1);
	}

	if (intmask > 0xffff)
	{
		printf("%s: interrupt mask 0x%x invalid\n", DRIVER, intmask);
		card->init = 0xffff;
		return (-1);
	}

	if (debmask > 0xffff)
	{
		printf("%s: debounce mask 0x%x invalid\n", DRIVER, debmask);
		card->init = 0xffff;
		return (-1);
	}

	if (pmask > 0xffff)
	{
		printf("%s: pulse select mask 0x%x invalid\n", DRIVER, pmask);
		card->init = 0xffff;
		return (-1);
	}

	card->invert[port]  = ((dir & INVERT_BIT) != 0) ? 1 : 0;
	card->intmask[port] = 0xffff;				/* initialise all interrupt mask */
	
	card->restore[port].LKC = 0x0000;
	card->restore[port].CSR = CSR_ENABLE;		/* enable input scanning */
	
/* debounce enable */

	if (debrate > 0)
	{
		card->debrate[port] = debrate;
			
		card->restore[port].CSR |= (debrate - 1) << DEBOUNCE_SHIFT;
		card->restore[port].DBR  = 0xffff;
	}
	
	card->restore[port].CSR |= scanrate << SCAN_SHIFT;
	card->restore[port].CSR |= (dir & 3) << DIRECTION_SHIFT;
	card->restore[port].CSR |= (clock & 0x3) << CLOCK_SHIFT;
	
	card->restore[port].PSR = 0x0000;
	card->restore[port].PPR = pwidth;

	if (card->debrate[port] > 0)
	{
		card->restore[port].DBR = debmask;
	}

	card->restore[port].PSR = pmask;
	card->intmask[port] = intmask;

	printf("%s:  dir: %s  scan: %sHz  debounce: %sHz  pulse: %s, intmask:%x, pmask:%x, debmask:%x \n", DRIVER, dirs[dir & 0x01], scans[scanrate], bounce[debrate - 1], widths[pwidth], intmask, pmask, debmask);
	
	return (0);
}

static long drvHy8506Init(void)
{
	struct cardinfo *card = cardlist;
	int status, i;

	while (card != NULL)
	{

	/* Don't continue if there was a fault with the configuration */

		if (card->init == 0xffff)
		{
			printf("%s: Error in configuration!\n", DRIVER);
			return (-1);
		}

	/* register the card in A16 address space */
	
		card->regs[0] = (struct ioregs *) ipmBaseAddr(card->carrier, card->ipslot, ipac_addrIO);
		card->regs[1] = (struct ioregs *) ((unsigned char*)ipmBaseAddr(card->carrier, card->ipslot, ipac_addrIO) + 0x20);
		card->regs[2] = (struct ioregs *) ((unsigned char*)ipmBaseAddr(card->carrier, card->ipslot, ipac_addrIO) + 0x40);
		if (card->regs == NULL)
		{
			printf("%s: ipmBaseAddr failed for card %d.\n", DRIVER, card->cardnum);
			return (-1);
		}
			
	/* check the ROM */
			
		status = checkprom(card->regs[0]);
	
	/* init interrupts */

		if (card->vector[0] > 0)
		{
			scanIoInit(&(card->ioscanpvt));
						
			if (ipmIntConnect(card->carrier, card->ipslot, card->vector[0], (void *)drvISR, card->cardnum) != 0)
			{
				printf("%s: ipmIntConnect failed.\n", DRIVER);
				return (-1);
			}
			else
			{
				for(i=0; i<3; i++)
				{
					card->restore[i].VEC  = card->vector[i];
					card->restore[i].IMR  = card->intmask[i];
					card->restore[i].CSR |= 0x4000;		/* IO Int Enable */
				}
	
				ipmIrqCmd(card->carrier, card->ipslot, 0, ipac_irqEnable);
			}
		}	

	/* reboot handler */

		short reset = 0x0000;			
		card->regs[0]->CSR  = reset;
		card->regs[1]->CSR  = reset;
		card->regs[2]->CSR  = reset;
				
		
		if (dumpcard(card) == 0)
		{											
			card->init = 1;
			printf("%s: card %d configured at ipmBaseAddr 0x%x\n", DRIVER, card->cardnum, (unsigned int) card->regs[0]);
		}

		card = card->next;
	}
	
	return (0);
}

/*
	reset the cards before reboot
*/


/******************* Return pointer to ioscanpvt   *****************/

int Hy8506getioscanpvt(const int cardnum, IOSCANPVT *scanpvt)
{
	struct cardinfo *card = cardlist;
	
	while ((card != NULL) && (card->cardnum != cardnum)) card = card->next;
	
	if (card == NULL)
	{
		printf("%s: getioscanpvt: cardnum %d not found!\n", DRIVER, cardnum);
		return (-1);
	}

	*scanpvt = card->ioscanpvt;
	
	return (0);
}


 epicsExportAddress(drvet, drvHy8506);

/* These two are for EPICS R3.13
 epicsRegisterFunction(Hy8506Configure);
 epicsRegisterFunction(Hy8506PortConfig);*/

 epicsRegisterFunction(Hy8506getioscanpvt);

/* Code for iocsh registration */

/* Hy8506Configure */
static const iocshArg Hy8506InitArg0 = {"cardNum", iocshArgInt};
static const iocshArg Hy8506InitArg1 = {"carrier", iocshArgInt};
static const iocshArg Hy8506InitArg2 = {"ipSlotNum", iocshArgInt};
static const iocshArg Hy8506InitArg3 = {"Interrupt", iocshArgInt};

static const iocshArg * const Hy8506InitArgs[] =  {&Hy8506InitArg0,
                                                   &Hy8506InitArg1,
                                                   &Hy8506InitArg2,
                                                   &Hy8506InitArg3};

static const iocshFuncDef initHy8506 = {"Hy8506Configure", 4, Hy8506InitArgs};
static void initHy8506CallFunc(const iocshArgBuf *args)
{
    Hy8506Configure(args[0].ival, args[1].ival, args[2].ival, args[3].ival);
}


/* Hy8506ExtraConfig */

static const iocshArg Hy8506PortArg0 = {"cardNum", iocshArgInt};
static const iocshArg Hy8506PortArg1 = {"Port", iocshArgInt};
static const iocshArg Hy8506PortArg2 = {"Debounce rate", iocshArgInt};
static const iocshArg Hy8506PortArg3 = {"Pulse With", iocshArgInt};
static const iocshArg Hy8506PortArg4 = {"Scan rate", iocshArgInt};
static const iocshArg Hy8506PortArg5 = {"Direction", iocshArgInt};
static const iocshArg Hy8506PortArg6 = {"clock rate", iocshArgInt};
static const iocshArg Hy8506PortArg7 = {"Debounce mask", iocshArgInt};
static const iocshArg Hy8506PortArg8 = {"Pulse mask", iocshArgInt};
static const iocshArg Hy8506PortArg9 = {"Interrupt mask", iocshArgInt};

static const iocshArg * const Hy8506PortArgs[] =  {&Hy8506PortArg0,
                                                   &Hy8506PortArg1,
                                                   &Hy8506PortArg2,
                                                   &Hy8506PortArg3,
                                                   &Hy8506PortArg4,
                                                   &Hy8506PortArg5,
                                                   &Hy8506PortArg6,
                                                   &Hy8506PortArg7,
                                                   &Hy8506PortArg8,
                                                   &Hy8506PortArg9};

static const iocshFuncDef configHy8506 = {"Hy8506PortConfig", 10, Hy8506PortArgs};
static void configHy8506CallFunc(const iocshArgBuf *args)
{
    Hy8506PortConfig(args[0].ival, args[1].ival, args[2].ival, args[3].ival, args[4].ival,
                   args[5].ival, args[6].ival, args[7].ival, args[8].ival, args[9].ival);
}


static const iocshArg Hy8506ReportArg0 = { "interest", iocshArgInt};
static const iocshArg * const Hy8506ReportArgs[1] = {&Hy8506ReportArg0};
static const iocshFuncDef Hy8506ReportFuncDef = {"drvHy8506Report",1,Hy8506ReportArgs};
static void Hy8506ReportCallFunc(const iocshArgBuf *args) {
    drvHy8506Report(args[0].ival);
}


static void Hy8506Register(void)
{
    iocshRegister(&initHy8506, initHy8506CallFunc);
    iocshRegister(&configHy8506, configHy8506CallFunc);
    iocshRegister(&Hy8506ReportFuncDef,Hy8506ReportCallFunc);
}

epicsExportRegistrar(Hy8506Register);

